@@ -1,5 +1,5 @@
 /*	canvas.c
-	Copyright (C) 2004-2016 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2004-2019 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -52,9 +52,11 @@ int clone_mode = TRUE, clone_x0 = -1, clone_y0 = -1,			// Clone settings
 	clone_dx0, clone_dy0;
 
 int recent_files;					// Current recent files setting
+char *recent_filenames[MAX_RECENT];			// Recent filenames themselves
 
 int	show_paste,					// Show contents of clipboard while pasting
 	col_reverse,					// Painting with right button
+	status_on[STATUS_ITEMS],			// Show status bar items?
 	text_paste,					// Are we pasting text?
 	canvas_image_centre,				// Are we centering the image?
 	chequers_optimize,				// Are we optimizing the chequers for speed?
@@ -1120,7 +1122,8 @@ void pressed_lasso(int cut)
 void update_menus()			// Update edit/undo menu
 {
 	static int memwarn;
-	int i, j, statemap;
+	unsigned int statemap;
+	int i, j;
 
 	if (mem_undo_fail && !memwarn) memwarn = alert_box(_("Warning"),
 		_("You have not allocated enough memory in the Preferences window to undo this action."),
@@ -1231,7 +1234,7 @@ void update_stuff(int flags)
 	if (flags & CF_PREFS)
 	{
 		update_undo_depth();	// If undo depth was changed
-		update_recent_files();
+		update_recent_files(FALSE);
 		init_status_bar();	// Takes care of all statusbar parts
 	}
 	if (flags & CF_MENU)
@@ -1476,7 +1479,9 @@ typedef struct {
 	int fmask, ftype, ftypes[NUM_FTYPES];
 	int need_save, need_anim, need_undo, need_icc, script;
 	int jpeg_c, png_c, tga_c, jp2_c, xtrans[3], xx[3], xy[3];
-	int tiff_m, lzma_c;
+	int tiff_m, lzma_c, zstd_c;
+	int webp_p, webp_q, webp_c;
+	int lbm_c, lbm_p;
 	int gif_delay, undo, icc;
 	int w, h;
 	/* Note: filename is in system encoding */
@@ -1635,15 +1640,18 @@ static void change_image_format(fselector_dd *dt, void **wdata, int what,
 	void **where)
 {
 	int ftype;
-	unsigned int flags;
+	unsigned int xflags;
 
 	if (!dt->need_save) return; // no use
 	cmd_read(where, dt);
 	ftype = dt->ftypes[dt->ftype];
-	flags = ftype != FT_TIFF ? file_formats[ftype].flags :
-		tiff_formats[dt->tiff_m].flags;
+	xflags = ftype != FT_TIFF ? file_formats[ftype].xflags :
+		tiff_formats[dt->tiff_m].xflags;
+	if (ftype == FT_WEBP) xflags |= dt->webp_p ? XF_COMPV8 : XF_COMPV8L;
+	if ((ftype == FT_LBM) && (dt->fmask & FF_IDX)) xflags |= XF_PBM;
+
 	/* Hide/show name/value widget pairs */
-	cmd_setv(wdata, (void *)flags, WDATA_ACTMAP);
+	cmd_setv(wdata, (void *)xflags, WDATA_ACTMAP);
 }
 
 int ftype_selector(int mask, char *ext, int def, char **names, int *ftypes)
@@ -1716,9 +1724,15 @@ void init_ls_settings(ls_settings *settings, void **wdata)
 	settings->jpeg_quality = jpeg_quality;
 	settings->png_compression = png_compression;
 	settings->lzma_preset = lzma_preset;
+	settings->zstd_level = zstd_level;
 	settings->tiff_type = -1; /* Use default */
 	settings->tga_RLE = tga_RLE;
 	settings->jp2_rate = jp2_rate;
+	settings->webp_preset = webp_preset;
+	settings->webp_quality = webp_quality;
+	settings->webp_compression = webp_compression;
+	settings->lbm_pack = lbm_pack;
+	settings->lbm_pbm = lbm_pbm;
 	settings->gif_delay = preserved_gif_delay;
 
 	/* Read in settings */
@@ -1731,9 +1745,15 @@ void init_ls_settings(ls_settings *settings, void **wdata)
 		settings->jpeg_quality = dt->jpeg_c;
 		settings->png_compression = dt->png_c;
 		settings->lzma_preset = dt->lzma_c;
+		settings->zstd_level = dt->zstd_c;
 		settings->tiff_type = dt->tiff_m;
 		settings->tga_RLE = dt->tga_c;
 		settings->jp2_rate = dt->jp2_c;
+		settings->webp_preset = dt->webp_p;
+		settings->webp_quality = dt->webp_q;
+		settings->webp_compression = dt->webp_c;
+		settings->lbm_pack = dt->lbm_c;
+		settings->lbm_pbm = dt->lbm_p;
 		settings->gif_delay = dt->gif_delay;
 
 		settings->mode = dt->mode;
@@ -1758,15 +1778,15 @@ void init_ls_settings(ls_settings *settings, void **wdata)
 static void store_ls_settings(ls_settings *settings)
 {
 	int ftype = settings->ftype, ttype = settings->tiff_type;
-	unsigned int fflags = (ftype == FT_TIFF) && (ttype >= 0) ?
-		tiff_formats[ttype].flags : file_formats[ftype].flags;
+	unsigned int xflags = (ftype == FT_TIFF) && (ttype >= 0) ?
+		tiff_formats[ttype].xflags : file_formats[ftype].xflags;
 
 	switch (settings->mode)
 	{
 	case FS_PNG_SAVE:
-		if (fflags & FF_TRANS)
+		if (xflags & XF_TRANS)
 			mem_set_trans(settings->xpm_trans);
-		if (fflags & FF_SPOT)
+		if (xflags & XF_SPOT)
 		{
 			mem_xbm_hot_x = settings->hot_x;
 			mem_xbm_hot_y = settings->hot_y;
@@ -1774,17 +1794,26 @@ static void store_ls_settings(ls_settings *settings)
 		// Fallthrough
 	case FS_CHANNEL_SAVE:
 	case FS_COMPOSITE_SAVE:
-		if (fflags & FF_COMPJ)
+		if (xflags & XF_COMPJ)
 			jpeg_quality = settings->jpeg_quality;
-		if (fflags & (FF_COMPZ | FF_COMPZT))
+		if (xflags & (XF_COMPZ | XF_COMPZT))
 			png_compression = settings->png_compression;
-		if (fflags & FF_COMPLZ)
+		if (xflags & XF_COMPLZ)
 			lzma_preset = settings->lzma_preset;
-		if (fflags & FF_COMPR)
+		if (xflags & XF_COMPZS)
+			zstd_level = settings->zstd_level;
+		if (xflags & XF_COMPR)
 			tga_RLE = settings->tga_RLE;
-		if (fflags & FF_COMPJ2)
+		if (xflags & XF_COMPJ2)
 			jp2_rate = settings->jp2_rate;
-		if ((fflags & FF_COMPT) && (ttype >= 0))
+		if (xflags & XF_COMPW)
+		{
+			webp_preset = settings->webp_preset;
+			if (webp_preset) webp_quality = settings->webp_quality;
+			else webp_compression = settings->webp_compression;
+		}
+		if (xflags & XF_COMPWT) webp_quality = settings->webp_quality;
+		if ((xflags & XF_COMPT) && (ttype >= 0))
 		{
 			/* Remember selection for incompatible types separately;
 			 * image is not set up yet, so use the mode instead */
@@ -1794,6 +1823,10 @@ static void store_ls_settings(ls_settings *settings)
 				mem_img_bpp == 3 ? &tiff_rtype :
 				mem_cols > 2 ? &tiff_itype : &tiff_btype) = ttype;
 		}
+		if (xflags & XF_COMPRL) lbm_pack = settings->lbm_pack;
+		/* !!! XF_PBM is only for indexed LBMs, but a complicated check
+		 * for that like for TIFFs above would add nothing of value here */
+		if (ftype == FT_LBM) lbm_pbm = settings->lbm_pbm;
 		break;
 	case FS_EXPORT_GIF:
 		preserved_gif_delay = settings->gif_delay;
@@ -1812,10 +1845,6 @@ static void fs_ok(fselector_dd *dt, void **wdata)
 	/* Pick up extra info */
 	init_ls_settings(&settings, wdata);
 
-	/* Needed to show progress in Windows GTK+2 */
-// !!! Disable for now, see what happens
-//	gtk_window_set_modal(GTK_WINDOW(GET_REAL_WINDOW(wdata)), FALSE);
-
 	/* Looks better if no dialog under progressbar */
 	cmd_showhide(GET_WINDOW(wdata), FALSE);
 
@@ -2009,8 +2038,6 @@ static void fs_ok(fselector_dd *dt, void **wdata)
 	if (redo && !script_cmds) /* Redo */
 	{
 		cmd_showhide(GET_WINDOW(wdata), TRUE);
-// !!! Disable for now, see what happens
-//		gtk_window_set_modal(GTK_WINDOW(GET_REAL_WINDOW(wdata)), TRUE);
 		return;
 	}
 	/* Done */
@@ -2027,27 +2054,37 @@ static void *fselector_code[] = {
 		OPTDe(ftnames, ftype, change_image_format), TRIGGER,
 	ENDIF(1),
 	IFx(need_save, 1),
-		VISMASK(~0),
-		MLABELr(_("Transparency index")), ACTMAP(FF_TRANS),
-			SPINa(xtrans), ACTMAP(FF_TRANS),
-		MLABELr(_("TIFF Compression")), ACTMAP(FF_COMPT),
-		OPTDe(tiffnames, tiff_m, change_image_format), ACTMAP(FF_COMPT),
-		MLABELr(_("JPEG Save Quality (100=High)")), ACTMAP(FF_COMPJ),
-			SPIN(jpeg_c, 0, 100), ACTMAP(FF_COMPJ),
-		MLABELr(_("ZIP Compression (0=None)")), ACTMAP(FF_COMPZT),
-		MLABELr(_("PNG Compression (0=None)")), ACTMAP(FF_COMPZ),
-			SPIN(png_c, 0, 9), ACTMAP(FF_COMPZ | FF_COMPZT),
+		VISMASK(~0), HEIGHTBAR,
+		MLABELr(_("Transparency index")), ACTMAP(XF_TRANS),
+			SPINa(xtrans), ACTMAP(XF_TRANS),
+		MLABELr(_("TIFF Compression")), ACTMAP(XF_COMPT),
+		OPTDe(tiffnames, tiff_m, change_image_format), ACTMAP(XF_COMPT),
+		MLABELr(_("JPEG Save Quality (100=High)")), ACTMAP(XF_COMPJ),
+			SPIN(jpeg_c, 0, 100), ACTMAP(XF_COMPJ),
+		MLABELr(_("ZIP Compression (0=None)")), ACTMAP(XF_COMPZT),
+		MLABELr(_("PNG Compression (0=None)")), ACTMAP(XF_COMPZ),
+			SPIN(png_c, 0, 9), ACTMAP(XF_COMPZ | XF_COMPZT),
 			ALTNAME("ZIP Compression (0=None)"),
-		MLABELr(_("LZMA2 Compression (0=None)")), ACTMAP(FF_COMPLZ),
-			SPIN(lzma_c, 0, 9), ACTMAP(FF_COMPLZ),
-		MLABELr(_("TGA RLE Compression")), ACTMAP(FF_COMPR),
-			SPIN(tga_c, 0, 1), ACTMAP(FF_COMPR),
-		MLABELr(_("JPEG2000 Compression (0=Lossless)")), ACTMAP(FF_COMPJ2),
-			SPIN(jp2_c, 0, 100), ACTMAP(FF_COMPJ2),
-		MLABELr(_("Hotspot at X =")), ACTMAP(FF_SPOT),
-			SPINa(xx), ACTMAP(FF_SPOT),
-		MLABELr(_("Y =")), ACTMAP(FF_SPOT),
-			SPINa(xy), ACTMAP(FF_SPOT),
+		MLABELr(_("LZMA2 Compression (0=None)")), ACTMAP(XF_COMPLZ),
+			SPIN(lzma_c, 0, 9), ACTMAP(XF_COMPLZ),
+		MLABELr(_("ZSTD Compression Level")), ACTMAP(XF_COMPZS),
+			SPIN(zstd_c, ZSTD_MIN, ZSTD_MAX), ACTMAP(XF_COMPZS),
+		CHECK(_("TGA RLE Compression"), tga_c), ACTMAP(XF_COMPR),
+		CHECK("PBM", lbm_p), ACTMAP(XF_PBM),
+		CHECK(_("LBM PackBits Compression"), lbm_c), ACTMAP(XF_COMPRL),
+		MLABELr(_("JPEG2000 Compression (0=Lossless)")), ACTMAP(XF_COMPJ2),
+			SPIN(jp2_c, 0, 100), ACTMAP(XF_COMPJ2),
+		MLABELr(_("WebP Compression")), ACTMAP(XF_COMPW),
+		OPTe(webp_presets, 0, webp_p, change_image_format), ACTMAP(XF_COMPW),
+		MLABELr(_("WebP Save Quality (100=High)")), ACTMAP(XF_COMPWT),
+		MLABELr(_("V8 Save Quality (100=High)")), ACTMAP(XF_COMPV8),
+			SPIN(webp_q, 0, 100), ACTMAP(XF_COMPV8 | XF_COMPWT),
+		MLABELr(_("V8L Compression (0=None)")), ACTMAP(XF_COMPV8L),
+			SPIN(webp_c, 0, 9), ACTMAP(XF_COMPV8L),
+		MLABELr(_("Hotspot at X =")), ACTMAP(XF_SPOT),
+			SPINa(xx), ACTMAP(XF_SPOT),
+		MLABELr(_("Y =")), ACTMAP(XF_SPOT),
+			SPINa(xy), ACTMAP(XF_SPOT),
 	ENDIF(1),
 	IFx(need_anim, 1),
 		MLABEL(_("Animation delay")),
@@ -2236,6 +2273,10 @@ void file_selector_x(int action_type, void **xdata)
 	tdata.tga_c = tga_RLE;
 	tdata.jp2_c = jp2_rate;
 	tdata.lzma_c = lzma_preset;
+	tdata.zstd_c = zstd_level;
+	tdata.webp_p = webp_preset;
+	tdata.webp_q = webp_quality;
+	tdata.webp_c = webp_compression;
 
 	tdata.gif_delay = preserved_gif_delay;
 	tdata.undo = undo_load;
@@ -3450,37 +3491,55 @@ void refresh_line(int mode, const int *lxy, rgbcontext *ctx)
 	if (line_clip(line, lxy, &j) >= 0) render_line(mode, line, j, ctx);
 }
 
-void update_recent_files()			// Update the menu items
+void update_recent_files(int save)		// Update the menu items
 {
 	char txt[64], *t, txt2[PATHTXT];
-	int i, count = 0;
+	int i, j;
 
-	for (i = 0; i < recent_files; i++)	// Display recent filenames
+	/* Init from inifile, skipping empty lines */
+	if (!recent_filenames[0])
 	{
-		sprintf(txt, "file%i", i + 1);
-
-		t = inifile_get(txt, "");
-		if (t[0])
+		for (i = j = 0; i < MAX_RECENT; i++)
 		{
-			gtkuncpy(txt2, t, PATHTXT);
-			cmd_setv(menu_slots[MENU_RECENT1 + i], txt2, LABEL_VALUE);
-			count++;
+			sprintf(txt, "file%i", i + 1);
+			t = inifile_get(txt, "");
+			if (t[0]) recent_filenames[j++] = strdup(t);
 		}
-		cmd_showhide(menu_slots[MENU_RECENT1 + i], t[0]); // Hide if empty
+		while (j < MAX_RECENT) recent_filenames[j++] = "";
 	}
-	for (; i < MAX_RECENT; i++)		// Hide extra items
-		cmd_showhide(menu_slots[MENU_RECENT1 + i], FALSE);
 
-	// Hide separator if not needed
-	cmd_showhide(menu_slots[MENU_RECENT_S], count);
+	if (save) /* Store to inifile */
+	{
+// !!! Or maybe set everything from recent_files to ""?
+		for (i = 0; i < MAX_RECENT; i++)
+		{
+			sprintf(txt, "file%i", i + 1);
+			inifile_set(txt, recent_filenames[i]);
+		}
+		return;
+	}
+
+	/* Display recent filenames */
+	for (i = 0; i < recent_files; i++)
+	{
+		if (!recent_filenames[i][0]) break; // Should only be at end
+		gtkuncpy(txt2, recent_filenames[i], PATHTXT);
+		cmd_setv(menu_slots[MENU_RECENT1 + i], txt2, LABEL_VALUE);
+		cmd_showhide(menu_slots[MENU_RECENT1 + i], TRUE); // Show
+	}
+	/* Hide separator if not needed */
+	cmd_showhide(menu_slots[MENU_RECENT_S], i);
+	/* Hide extra items */
+	for (; i < MAX_RECENT; i++) cmd_showhide(menu_slots[MENU_RECENT1 + i], FALSE);
 }
 
-void register_file( char *filename )		// Called after successful load/save
+void register_file(char *filename)	// Called after successful load/save
 {
-	char txt[64], txt1[64], *c;
-	int i, f;
+	char *c;
+	int i;
 
-	c = strrchr( filename, DIR_SEP );
+	if (!filename[0]) return;	// Paranoia
+	c = strrchr(filename, DIR_SEP);
 	if (c)
 	{
 		i = *c;
@@ -3489,30 +3548,25 @@ void register_file( char *filename )		// Called after successful load/save
 		*c = i;
 	}
 
-	// Is it already in used file list?  If so shift relevant filenames down and put at top.
-	i = 1;
-	f = 0;
-	while ( i<MAX_RECENT && f==0 )
+	/* Is it already in used file list? */
+	for (i = 0; i < MAX_RECENT; i++)
+		if (!strcmp(recent_filenames[i], filename)) break;
+	/* Copy it into LAST slot if not */
+	if (i >= MAX_RECENT)
 	{
-		sprintf( txt, "file%i", i );
-		c = inifile_get(txt, "");
-		if ( strcmp( filename, c ) == 0 ) f = 1;	// Filename found in list
-		else i++;
+		i = MAX_RECENT - 1;
+		if (recent_filenames[i][0]) free(recent_filenames[i]);
+		recent_filenames[i] = strdup(filename);
 	}
-	if ( i>1 )			// If file is already most recent, do nothing
+	/* Now rotate it to the top if not already there */
+	if (i > 0)
 	{
-		while ( i>1 )
-		{
-			sprintf( txt, "file%i", i-1 );
-			sprintf( txt1, "file%i", i );
-			inifile_set(txt1, inifile_get(txt, ""));
-
-			i--;
-		}
-		inifile_set("file1", filename);		// Strip off filename
+		filename = recent_filenames[i];
+		memmove(recent_filenames + 1, recent_filenames,
+			i * sizeof(recent_filenames[0]));
+		recent_filenames[0] = filename;
+		update_recent_files(FALSE);
 	}
-
-	update_recent_files();
 }
 
 void create_default_image()			// Create default new image
