@@ -1,5 +1,5 @@
 /*	fpick.c
-	Copyright (C) 2007-2015 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2007-2019 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -355,6 +355,8 @@ static void fpick_scan_drives(fpick_dd *dt)	// Scan drives, populate widgets
 
 #endif
 
+#define MAX_DIR_FILES (16 * 1024 * 1024) /* 16+ million is when to say "Enough" */
+
 static void scan_dir(fpick_dd *dt, DIR *dp, char *select)
 {
 	char full_name[PATHBUF], txt_size[64], txt_date[64], tmp_txt[64];
@@ -382,7 +384,7 @@ static void scan_dir(fpick_dd *dt, DIR *dp, char *select)
 		cnt++;
 	}
 
-	while ((ep = readdir(dp)))
+	while ((cnt < MAX_DIR_FILES) && (ep = readdir(dp)))
 	{
 		wjstrcat(full_name, PATHBUF, dir, l, ep->d_name, NULL);
 
@@ -686,7 +688,8 @@ static void fpick_file_dialog(fpick_dd *dt, void **wdata, int what, void **where
 		if (fname[1] == ':') return; // Drive
 #endif
 
-		sprintf(tdata.title = fnm, "%s / %s", __("Delete"), __("Rename"));
+		snprintf(tdata.title = fnm, sizeof(fnm),
+			"%s / %s", __("Delete"), __("Rename"));
 		tdata.what = _("Enter the new filename");
 		strncpy0(tdata.fname, fname, PATHBUF);
 	}
@@ -1010,7 +1013,6 @@ GtkWidget *fpick(GtkWidget **box, char *title, int flags, void **r)
 
 	fp = gtk_file_selection_new(__(title));
 	fs = GTK_FILE_SELECTION(fp);
-	gtk_window_set_modal(GTK_WINDOW(fp), TRUE);
 	if ( flags & FPICK_DIRS_ONLY )
 	{
 		gtk_widget_hide(GTK_WIDGET(fs->selection_entry));
