@@ -1,5 +1,5 @@
 /*	memory.h
-	Copyright (C) 2004-2016 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2004-2019 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -176,7 +176,7 @@ typedef struct {
 	int channel;			// Current active channel
 	int ics;			// Has the centre been set by the user?
 	float ic[2];			// Current centre x,y
-	int tool_pat;			// Tool pattern number
+	int tool_pat, tool_pat_B;	// Tool pattern number
 	int xbm_hot_x, xbm_hot_y;	// Current XBM hot spot
 	char prot_mask[256];		// 256 bytes used for indexed images
 	int prot;			// Number of protected colours in prot_RGB
@@ -280,7 +280,8 @@ enum {
 	BLEND_SATPP,
 
 	/* Per-channel modes */
-	BLEND_MULT,
+	BLEND_1BPP,	// First one-byte mode
+	BLEND_MULT = BLEND_1BPP,
 	BLEND_DIV,
 	BLEND_SCREEN,
 // !!! No "Overlay" - it is a reverse "Hard Light"
@@ -293,10 +294,10 @@ enum {
 	BLEND_LIGHT,
 	BLEND_GRAINX,
 	BLEND_GRAINM,
+	BLEND_XHOLD,
 
 	BLEND_NMODES
 };
-#define BLEND_1BPP BLEND_MULT /* First one-byte mode */
 
 #define BLEND_MMASK    0x3F
 #define BLEND_XFORM    0x40
@@ -376,6 +377,7 @@ image_state mem_state;			// Current edit settings
 #define mem_icy			mem_state.ic[1]
 #define mem_ics			mem_state.ics
 #define mem_tool_pat		mem_state.tool_pat
+#define mem_tool_pat_B		mem_state.tool_pat_B
 #define mem_xbm_hot_x		mem_state.xbm_hot_x
 #define mem_xbm_hot_y		mem_state.xbm_hot_y
 #define mem_prot_mask		mem_state.prot_mask
@@ -424,9 +426,28 @@ transform_state mem_bcsp[2];
 #define BRCOSA_POSTERIZE 3
 #define DEF_POSTERIZE(T) ((T) ? 256 : 8) /* truncated/rounded vs bitwise */
 
+/// THRESHOLD
+
+enum {
+	XHOLD_MAX = 0,
+	XHOLD_MIN = 1,
+	XHOLD_RED,
+	XHOLD_GREEN,
+	XHOLD_BLUE,
+
+	XHOLD_NMODES
+};
+
+typedef struct {
+	int lo, hi, mode;
+} threshold_state;
+
+threshold_state mem_ts;
+
 /// PATTERNS
 
-unsigned char mem_pattern[8 * 8];	// Current pattern
+int pattern_B;				// Let colour B have its own pattern
+unsigned char *mem_pattern;		// Current pattern
 unsigned char mem_col_pat[8 * 8];	// Indexed 8x8 colourised pattern using colours A & B
 unsigned char mem_col_pat24[8 * 8 * 3];	// RGB 8x8 colourised pattern using colours A & B
 
@@ -486,15 +507,12 @@ static inline unsigned int bitcount(unsigned int n)
 	n -= (n >> 1) & 0x5555555555555555ULL;
 	m = n & 0xCCCCCCCCCCCCCCCCULL; n = (n ^ m) + (m >> 2);
 	n = (n + (n >> 4)) & 0x0F0F0F0F0F0F0F0FULL;
-	n = (n + (n >> 8)) & 0x00FF00FF00FF00FFULL;
-	n = (n + (n >> 16)) & 0x0000FFFF0000FFFFULL;
-	n = (n + (n >> 32)) & 0x00000000FFFFFFFFULL;
+	n = (n * 0x0101010101010101ULL) >> 56;
 #else
 	n -= (n >> 1) & 0x55555555;
 	m = n & 0x33333333; n = m + ((n ^ m) >> 2);
 	n = (n + (n >> 4)) & 0x0F0F0F0F;
-	n = (n + (n >> 8)) & 0x00FF00FF;
-	n = (n + (n >> 16)) & 0x0000FFFF;
+	n = (n * 0x01010101) >> 24;
 #endif
 	return (n);
 }
@@ -798,6 +816,10 @@ void mem_smudge(int ox, int oy, int nx, int ny);
 void do_transform(int start, int step, int cnt, unsigned char *mask,
 	unsigned char *imgr, unsigned char *img0, int m0);
 
+//	Apply thresholding
+void do_xhold(int start, int step, int cnt, unsigned char *mask,
+	unsigned char *imgr, unsigned char *img0);
+
 void mem_flip_v(char *mem, char *tmp, int w, int h, int bpp);	// Flip image vertically
 void mem_flip_h( char *mem, int w, int h, int bpp );		// Flip image horizontally
 int mem_sel_rot( int dir );					// Rotate clipboard 90 degrees
