@@ -1,5 +1,5 @@
 /*	otherwindow.c
-	Copyright (C) 2004-2016 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2004-2019 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -384,12 +384,15 @@ static int click_pat(pattern_dd *dt, void **wdata, int what, void **where,
 		mem_col_24[ab] = mem_pal[pat_no];
 		update_stuff(UPD_AB);
 	}
-	else if (mouse->button != 1) return (FALSE); // Left click only
 	else if (dt->mode == CHOOSE_PATTERN)
 	{
-		mem_tool_pat = pat_no;
+		if (mouse->button == 1) mem_tool_pat = pat_no; // Left
+		else if (pattern_B && (mouse->button == 3)) // Right
+			mem_tool_pat_B = pat_no;
+		else return (FALSE);
 		update_stuff(UPD_PAT);
 	}
+	else if (mouse->button != 1) return (FALSE); // Left click only
 	else /* if (dt->mode == CHOOSE_BRUSH) */
 	{
 		mem_set_brush(pat_no);
@@ -425,11 +428,11 @@ void choose_pattern(int typ)	// Bring up pattern chooser (0) or brush (1)
 	tdata.wh[2] = 3; // bpp
 	if (typ == CHOOSE_PATTERN)
 	{
-		tdata.wh[0] = PATTERN_GRID_W * (8 * 4 + 4);
-		tdata.wh[1] = PATTERN_GRID_H * (8 * 4 + 4);
-		tdata.xs = tdata.ys = 8 * 4 + 4;
-		tdata.xw = PATTERN_GRID_W;
-		tdata.max = PATTERN_GRID_W * PATTERN_GRID_H - 1;
+		tdata.wh[0] = patterns_grid_w * PATTERN_CELL;
+		tdata.wh[1] = patterns_grid_h * PATTERN_CELL;
+		tdata.xs = tdata.ys = PATTERN_CELL;
+		tdata.xw = patterns_grid_w;
+		tdata.max = patterns_grid_w * patterns_grid_h - 1;
 	}
 	else if (typ == CHOOSE_BRUSH)
 	{
@@ -2952,3 +2955,111 @@ void pressed_segment()
 
 	run_create_(seg_code, &tdata, sizeof(tdata), script_cmds);
 }
+
+/// THRESHOLDING WINDOW
+
+int xhold_preview;
+
+typedef struct {
+	int lo[3], hi[3], mode, preview;
+	int rgb;
+	void **lspin, **hspin, **pbutton;
+} xhold_dd;
+
+static void xhold_changed(xhold_dd *dt, void **wdata, int what, void **where)
+{
+	void *cause = cmd_read(where, dt);
+
+	if (dt->lo[0] > dt->hi[0]) // Inversion tried to happen
+	{
+		if (cause == dt->hspin) cmd_set(dt->lspin, dt->hi[0]);
+		else cmd_set(dt->hspin, dt->lo[0]);
+		return; // Changed slider did the update
+	}
+	mem_ts.lo = dt->lo[0];
+	mem_ts.hi = dt->hi[0];
+	mem_ts.mode = dt->mode;
+	if (xhold_preview) update_stuff(UPD_RENDER);
+}
+
+static void xhold_evt(xhold_dd *dt, void **wdata, int what, void **where)
+{
+	int update = 0;
+
+	if (what == op_EVT_CHANGE) // Toggle preview
+	{
+		cmd_read(where, dt);
+		if (dt->preview ^ !xhold_preview) return; // Nothing to do
+		xhold_preview = dt->preview;
+		update_stuff(UPD_RENDER);
+		return;
+	}
+
+	/* First, disable preview */
+	if (xhold_preview) update = UPD_RENDER;
+	xhold_preview = FALSE;
+
+	while (what == op_EVT_OK)
+	{
+		unsigned char *dest, *xbuf, *mask;
+		int i, bpp = MEM_BPP;
+
+		mask = malloc(mem_width * (bpp + 1));
+		if (!mask) break;
+		spot_undo(UNDO_FILT);
+		xbuf = mask + mem_width;
+
+		for (i = 0; i < mem_height; i++)
+		{
+			row_protected(0, i, mem_width, mask);
+			dest = mem_img[mem_channel] + i * mem_width * bpp;
+			do_xhold(0, 1, mem_width, mask, xbuf, dest);
+			process_img(0, 1, mem_width, mask, dest, dest, xbuf,
+				NULL, bpp, BLENDF_SET | BLENDF_INVM);
+		}
+
+		free(mask);
+		mem_undo_prepare();
+		update |= UPD_IMG;
+		break;
+	}
+
+	run_destroy(wdata); // Finished
+	update_stuff(update);
+}
+
+static char *xh_txt[] = {_("Max"), _("Min"), _("Red"), _("Green"), _("Blue")};
+
+#define WBbase xhold_dd
+static void *xhold_code[] = {
+	WINDOWm(_("Threshold Image")), DEFW(300),
+	TABLE2(2), OPNAME0,
+	REF(lspin), TSPINSLIDEa(_("From"), lo),
+		EVENT(CHANGE, xhold_changed), TRIGGER,
+	REF(hspin), TSPINSLIDEa(_("To"), hi), EVENT(CHANGE, xhold_changed),
+	WDONE,
+	IFx(rgb, 1),
+		BORDER(RPACK, 0),
+		RPACKe(xh_txt, XHOLD_NMODES, 0, mode, xhold_changed),
+		OPNAME(""), // default
+	ENDIF(1),
+	HSEP,
+	EQBOX,
+	CANCELBTN(_("Cancel"), xhold_evt),
+	REF(pbutton), TOGGLE(_("Preview"), preview, xhold_evt),
+	OKBTN(_("OK"), xhold_evt),
+	WSHOW
+};
+#undef WBbase
+
+void pressed_xhold()
+{
+	xhold_dd tdata = { { 128, 0, 255 }, { 255, 0, 255 }, 0, FALSE, MEM_BPP == 3 };
+
+	if (IS_INDEXED)
+	{
+		tdata.lo[0] = mem_cols / 2;
+		tdata.hi[0] = tdata.hi[2] = tdata.lo[2] = mem_cols - 1;
+	}
+	run_create_(xhold_code, &tdata, sizeof(tdata), script_cmds);
+}
