@@ -1,5 +1,5 @@
 /*	vcode.h
-	Copyright (C) 2013-2016 Dmitry Groshev
+	Copyright (C) 2013-2019 Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -149,6 +149,7 @@ enum {
 
 	op_MOUNT,
 	op_REMOUNT,
+	op_HEIGHTBAR,
 	op_uCHECK,
 	op_uCHECKb,
 	op_uSPIN,
@@ -256,7 +257,7 @@ enum {
 	op_MKSHRINK,
 	op_NORESIZE,
 	op_WANTMAX,
-	op_KEEPWIDTH,
+	op_KEEPSIZE,
 	op_WXYWH,
 	op_WPMOUSE,
 	op_WPWHEREVER,
@@ -373,7 +374,7 @@ typedef struct {
 //	Structure which is supplied to WINDOW_TEXTENG
 typedef struct {
 	char *text, *font;
-	int angle, dpi, align;
+	int angle, dpi, align, spacing;
 	rgbcontext ctx;
 } texteng_dd;
 
@@ -413,6 +414,7 @@ typedef int (*evtxr_fn)(void *ddata, void **wdata, int what, void **where,
 //	Textengine flags
 int texteng_aa;  /* Antialias */
 int texteng_rot; /* Rotate */
+int texteng_spc; /* Set spacing */
 int texteng_dpi; /* Set DPI */
 int texteng_lf;  /* Handle newlines */
 int texteng_con; /* Work in console mode */
@@ -673,6 +675,7 @@ enum {
 #define XHBOXS XHBOXbp(5, 0, 0)
 #define XHBOXBS XHBOXbp(5, 5, 0)
 #define TLHBOXl(X,Y,L) WBh_t(HBOX, 1), WBxyl(X, Y, L)
+#define TLHBOXpl(P,X,Y,L) WBh_t(HBOX, 2), WBpbs(P, 0, 0), WBxyl(X, Y, L)
 #define FVBOX(NM) FRAME(NM), VBOX
 #define FVBOXB(NM) FRAME(NM), VBOXB
 #define FVBOXBS(NM) FRAME(NM), VBOXBS
@@ -794,6 +797,7 @@ enum {
 #define T1SPINSLIDE(V,V0,V1) WBrhf_t1(SPINSLIDE, 4), WBfield(V), \
 	(void *)(V0), (void *)(V1), WBwh(255, 20)
 #define TSPINSLIDE(NM,V,V0,V1) T1SPINSLIDE(V,V0,V1), TLABEL(NM)
+#define TSPINSLIDEa(NM,A) WBrhf_t1(SPINSLIDEa, 1), WBfield(A), TLABEL(NM)
 #define TLSPINSLIDE(V,V0,V1,X,Y) WBrhf_t(SPINSLIDE, 4), WBfield(V), \
 	(void *)(V0), (void *)(V1), WBxyl(X, Y, 1)
 #define TLSPINSLIDEvs(V,V0,V1,X,Y) WBrh_t(SPINSLIDE, 5), &(V), \
@@ -811,6 +815,7 @@ enum {
 #define TLCHECK(NM,V,X,Y) TLCHECKl(NM, V, X, Y, 1)
 #define TLCHECKvl(NM,V,X,Y,L) WBrh_t(CHECK, 3), &(V), (NM), WBxyl(X, Y, L)
 #define TLCHECKv(NM,V,X,Y) TLCHECKvl(NM, V, X, Y, 1)
+#define uCHECK(NM,V) WBrhf_(uCHECK, 2), WBfield(V), (NM)
 #define uCHECKv(NM,V) WBrh_(uCHECK, 2), &(V), (NM)
 /* !!! No more than 255 choices */
 #define RPACK(SS,N,H,V) WBrhf_x(RPACK, 3), WBfield(V), (SS), WBnh(N, H)
@@ -838,6 +843,8 @@ enum {
 /* !!! These blocks each hold 1 nested EVENT block */
 #define OPTve(SS,N,V,HS) WBr2h_(OPT, 3 + 2), &(V), (SS), (void *)(N), \
 	EVENT(SELECT, HS)
+#define OPTe(SS,N,V,HS) WBr2hf_(OPT, 3 + 2), WBfield(V), (SS), (void *)(N), \
+	EVENT(SELECT, HS)
 #define XOPTe(SS,N,V,HS) WBr2hf_x(OPT, 3 + 2), WBfield(V), (SS), (void *)(N), \
 	EVENT(SELECT, HS)
 #define TLOPTle(SS,N,V,HS,X,Y,L) WBr2hf_t(OPT, 4 + 2), WBfield(V), (SS), \
@@ -868,6 +875,7 @@ enum {
 #define LISTCX(V,L,SM,M,HS,HX) WBr3hf_(LISTCX, 4 + 2 * 2), WBfield(V), \
 	WBfield(L), WBfield(SM), WBfield(M), EVENT(SELECT, HS), EVENT(EXT, HX)
 #define XENTRY(V) WBrhf_x(ENTRY, 1), WBfield(V)
+#define XLENTRY(V,MX) WBrhf_x(ENTRY, 2), WBfield(V), (void *)(MX)
 #define MLENTRY(V) WBrhf_(MLENTRY, 1), WBfield(V)
 #define TLENTRY(V,MX,X,Y,L) WBrhf_t(ENTRY, 3), WBfield(V), (void *)(MX), \
 	WBxyl(X, Y, L)
@@ -971,6 +979,7 @@ enum {
 #define PMOUNT(V,FN,H,K,NK) WBr2hf_x(MOUNT, 4 + 2), WBfield(V), (FN), (K), \
 	(void *)(NK), EVENT(CHANGE, H)
 #define REMOUNTv(V) WBrh_x(REMOUNT, 1), &(V)
+#define HEIGHTBAR WBh_e(HEIGHTBAR, 0)
 //#define EXEC(FN) WBh(EXEC, 1), (FN)
 #define GOTO(A) WBh(GOTO, 1), (A)
 #define CALL(A) WBh(CALL, 1), (A)
@@ -1018,7 +1027,8 @@ enum {
 #define FOCUS WBh(FOCUS, 0)
 #define WIDTH(N) WBh(WIDTH, 1), (void *)(N)
 #define MINWIDTH(N) WBh(WIDTH, 1), (void *)(-(N))
-#define KEEPWIDTH WBh(KEEPWIDTH, 0)
+#define KEEPWIDTH WBh(KEEPSIZE, 0)
+#define KEEPHEIGHT WBh(KEEPSIZE, 1), NULL
 #define HEIGHT(N) WBh(HEIGHT, 1), (void *)(N)
 #define ONTOP(V) WBhf(ONTOP, 1), WBfield(V)
 #define ONTOP0 WBh(ONTOP, 0)
